/*
     eib.c - Low level EIB driver for BCU 1

     Copyright (C) 2000 Bernd Thallner <bernd@kangaroo.at>

     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


     2001-11-11: Driver is able to compile and run in 2.4 kernel 
                 compatibility mode

*/

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/serial_reg.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/wait.h>

#include <asm/uaccess.h>
#include <asm/io.h>

#include"eib.h"

EXPORT_NO_SYMBOLS;

static char *eib_name="Low level EIB driver for BCU 1";
static char *eib_version="0.2.2";

// eib bcu version
#define EIB_BCU_VERSION 1

// buffer size
#define MAX_READ_BUFFER 20 // * 33 byte
#define MAX_WRITE_BUFFER 5 // * 33 byte

// maximum time in usec to wait between rx interrupt raised and byte fully send
#define MAX_TEMT_WAIT 1000

// minimum time in usec to wait between byte fully send and clear rts signal
#define WRITE_TEMT_CLRRTS_DELAY 100

// minimum time in usec to wait between clear rts and set rts signal
#define WRITE_CLRRTS_SETRTS_DELAY 80

// timeout for data block transfer
#define READ_WRITE_TIMEOUT 130000

// maximum time between to interrupts if not in idle state
#define MAX_TIME_BETWEEN_INT 80000

// minimum time to wait after reset without any respond to signals
#define MIN_RESET_WAIT_TIME 130000

// enable debug messsages
//#define EIB_DEBUG

// sniff_falcon_hack
// #define SNIFF_HACK

#if !defined(DEFAULT_IO)
  #define DEFAULT_IO 0x02f8
#endif

#if !defined(DEFAULT_IRQ)
  #define DEFAULT_IRQ 3
#endif

#if !defined(DEFAULT_MAJOR)
  #define DEFAULT_MAJOR 0
#endif

static ssize_t eib_read(struct file *file, char *buf, size_t count, loff_t *offset);
static ssize_t eib_write(struct file *file, const char *buf, size_t count, loff_t *offset);
static unsigned int eib_poll(struct file *file, poll_table *wait);
static int eib_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static int eib_open(struct inode *inode, struct file *file);
static int eib_flush(struct file *file);
static int eib_release(struct inode *inode, struct file *file);

void interrupt_handler(int irq, void *dev_id, struct pt_regs *regs);
void initate_send_timer(unsigned long data);
void detect_reset_timer(unsigned long data);
void do_temt_bh(void);

#define setRTS() outb(UART_MCR_DTR | UART_MCR_RTS | UART_MCR_OUT2, io+UART_MCR)
#define clrRTS() outb(UART_MCR_DTR | UART_MCR_OUT2, io+UART_MCR)

#define getCTS() (inb(io+UART_MSR)&UART_MSR_CTS)

#define add_initate_send_timer()  { del_timer(&initate_send_timer_list); \
                                    init_timer(&initate_send_timer_list); \
                                    initate_send_timer_list.expires=jiffies+1; \
                                    initate_send_timer_list.function= \
                                    initate_send_timer; \
                                    add_timer(&initate_send_timer_list); \
                                  }

#define add_detect_reset_timer()  { del_timer(&detect_reset_timer_list); \
                                    init_timer(&detect_reset_timer_list); \
                                    detect_reset_timer_list.expires=jiffies+1; \
                                    detect_reset_timer_list.function= \
                                    detect_reset_timer; \
                                    add_timer(&detect_reset_timer_list); \
                                  }


struct file_operations fops = 
{
    read:       eib_read,
    write:      eib_write,
    poll:       eib_poll,
    ioctl:      eib_ioctl,
    open:       eib_open,
    flush:      eib_flush,
    release:    eib_release,
};


int io=DEFAULT_IO, irq=DEFAULT_IRQ, major=DEFAULT_MAJOR;


MODULE_AUTHOR("Bernd Thallner, bernd@kangaroo.at");
MODULE_DESCRIPTION("A low level EIB driver for BCU 1.");
MODULE_PARM(io, "i");
MODULE_PARM_DESC(io, "I/O port for the serial port where the BCU 1 is connected.");
MODULE_PARM(irq, "i");
MODULE_PARM_DESC(irq, "Interrupt number for the serial port where the BCU 1 is connected.");
MODULE_PARM(major, "i");
MODULE_PARM_DESC(major, "Major number for this device.");

#ifdef SNIFF_HACK
  int ignore_read_forced=0;
  MODULE_PARM(ignore_read_forced, "i");
#endif

unsigned long warning_flag=0;
unsigned char read_buffer[MAX_READ_BUFFER][33], write_buffer[MAX_WRITE_BUFFER][33];
int rb_col=0, wb_col=0;
volatile int state=0, rb_head=0, rb_length=0, wb_tail=0, wb_length=0, initate_send=0, eib_read_count=0, eib_write_count=0, eib_single_open=0, eib_open_count=0;
struct timeval last_int_tv, start_tv, reset_tv;
struct timer_list initate_send_timer_list, detect_reset_timer_list;
struct eib_device_information eib_dev_info;

/*struct wait_queue_head *eib_read_queue=NULL, *eib_write_queue=NULL;*/
/*declaration changed in Kernel 2.4 */
DECLARE_WAIT_QUEUE_HEAD(eib_read_queue);
DECLARE_WAIT_QUEUE_HEAD(eib_write_queue);
unsigned char old_dll, old_dlh, old_lcr, old_mcr;

int bitcount(unsigned char x) {
  int b;
 
  for(b=0; x!=0; x>>=1)
    if(x&01)
      b++;
 
  return b;
}

inline long time_difference(struct timeval *tv1, struct timeval *tv2) {
  struct timeval tv3;

  if(tv2==NULL) {
    do_gettimeofday(&tv3);
    tv2=&tv3;
  }
  return (tv2->tv_usec-tv1->tv_usec)+1000000*(tv2->tv_sec-tv1->tv_sec);
}

void do_reset(unsigned long reason) {

#ifdef EIB_DEBUG
  printk(KERN_DEBUG "debug: do_reset: state=%d, reason=%lx, initate_send=%d, getCTS()=%d, MCR=%x\n", state, reason, initate_send, getCTS(), inb(io+UART_MCR));
#endif

  warning_flag|=reason;
  state=7;
  inb(io+UART_IIR);
  inb(io+UART_RX);
  inb(io+UART_MSR);
  do_gettimeofday(&reset_tv);
  clrRTS();
}

void init_serial() {

  outb(0x00,                         io+UART_IER);
  old_lcr=inb(                       io+UART_LCR);
  outb(UART_LCR_DLAB,                io+UART_LCR);
  old_dll=inb(                       io+UART_DLL);
  outb(0x0c,                         io+UART_DLL);
  old_dlh=inb(                       io+UART_DLM);
  outb(0x00,                         io+UART_DLM);
  outb(UART_LCR_WLEN8,               io+UART_LCR);
  outb(0x00,                         io+UART_FCR);
  old_mcr=inb(                       io+UART_MCR);
  outb(UART_MCR_DTR | UART_MCR_OUT2, io+UART_MCR);
  outb(UART_IER_MSI | UART_IER_RDI,  io+UART_IER);
  inb(                               io+UART_RX);
}

void restore_serial() {
  outb(0x00,          io+UART_IER);
  outb(UART_LCR_DLAB, io+UART_LCR);
  outb(old_dll,       io+UART_DLL);
  outb(old_dlh,       io+UART_DLM);
  outb(old_lcr,       io+UART_LCR);
  outb(old_mcr,       io+UART_MCR);
}

static ssize_t eib_read(struct file *file, char *buf, size_t count, loff_t *offset) {
  int rb_tail, len;
  unsigned long flags;
  
  while(!rb_length || eib_read_count) 
    { // read buffer empty
      if(file->f_flags & O_NONBLOCK)
	return -EAGAIN;
      interruptible_sleep_on(&eib_read_queue);
      if(signal_pending(current))
	return -ERESTARTSYS;
    }
  
  save_flags(flags); cli();
  
  if(eib_read_count) 
    {
      restore_flags(flags);
      return -EAGAIN;
    }

  eib_read_count++;
  rb_tail=(rb_head-rb_length+MAX_READ_BUFFER)%MAX_READ_BUFFER;
  len=(read_buffer[rb_tail][0]&0x1f);
  
  restore_flags(flags);
  
  if(count<len) return -ENOMSG;
  if(copy_to_user(buf, &read_buffer[rb_tail][1], len))
    return -EFAULT;

  save_flags(flags); cli();
  rb_length--;
  eib_read_count=0;
  restore_flags(flags);
  return len;
}

static ssize_t eib_write(struct file *file, const char *buf, size_t count, loff_t *offset) {
  int temp, wb_head;
  unsigned long flags;

  if(count>32)
    return -EINVAL;
  while(wb_length==MAX_WRITE_BUFFER || eib_write_count) { // write buffer full
    if(file->f_flags & O_NONBLOCK)
      return -EAGAIN;
    interruptible_sleep_on(&eib_write_queue);
    if(signal_pending(current))
      return -ERESTARTSYS;
  }
  save_flags(flags); cli();
  if(eib_write_count) {
    restore_flags(flags);
    return -EAGAIN;
  }
  eib_write_count++;
  wb_head=(wb_tail+wb_length)%MAX_WRITE_BUFFER;
  restore_flags(flags);
  temp=(count&0x1f)|0x20;
  write_buffer[wb_head][0]=temp|(bitcount(temp)%2)*0x80;
  if(copy_from_user(&write_buffer[wb_head][1], buf, count))
    return -EFAULT;
  save_flags(flags); cli();
  wb_length++;
  add_initate_send_timer();
  eib_write_count=0;
  restore_flags(flags);
  return count&0x1f;
}

static unsigned int eib_poll(struct file *file, poll_table *wait) {
  unsigned int mask=0;

  poll_wait(file, &eib_read_queue, wait);
  poll_wait(file, &eib_write_queue, wait);
  if(rb_length)
    mask|=POLLIN|POLLRDNORM;
  if(wb_length!=MAX_WRITE_BUFFER)
    mask|=POLLOUT|POLLWRNORM;
  return mask;
}

static int eib_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg) {
  unsigned long flags;

  switch(cmd) {
    
  case EIB_GET_BCU_VERSION:
    return EIB_BCU_VERSION;
    
  case EIB_GET_DEVICE_INFORMATION:
    eib_dev_info.warning_flag=warning_flag;
    eib_dev_info.rb_length=rb_length;
    eib_dev_info.wb_length=wb_length;
    if(copy_to_user((void *)arg, (void *)&eib_dev_info, sizeof(struct eib_device_information)))
      return -EFAULT;
    break;

  case EIB_RESET_DEVICE_STATISTIC:
    save_flags(flags); cli();
    warning_flag=0;
    eib_dev_info.stat_rx_bytes=0;
    eib_dev_info.stat_tx_bytes=0;
    eib_dev_info.stat_rx_data_blocks=0;
    eib_dev_info.stat_tx_data_blocks=0;
    eib_dev_info.stat_max_rx_time_data_block=0;
    eib_dev_info.stat_max_tx_time_data_block=0;
    restore_flags(flags);
    break;

  case EIB_CLEAR_WRITE_BUFFER:
    save_flags(flags); cli();
    del_timer(&initate_send_timer_list);
    add_detect_reset_timer();
    initate_send=0;
    warning_flag=0;
    wb_length=0;
    do_reset(0x00400000);
    restore_flags(flags);
    wake_up_interruptible(&eib_write_queue);
    break;

  case EIB_SINGLE_OPEN:
    if(eib_open_count>1)
      return eib_open_count;
    eib_single_open=1;
    break;

  default:
    return -ENOIOCTLCMD;
  }
  return 0;
}

static int eib_open(struct inode *inode, struct file *file) {

  if(eib_single_open)
    return -EBUSY;
  MOD_INC_USE_COUNT;
  eib_open_count++;
  return 0;
}

static int eib_flush(struct file *file) {
  
  while(wb_length) { // write buffer not empty
    interruptible_sleep_on(&eib_write_queue);
    if(signal_pending(current))
      return -ERESTARTSYS;
  }
  return 0;
}

static int eib_release(struct inode *inode, struct file *file) {

  eib_open_count--;
  eib_single_open=0;
  MOD_DEC_USE_COUNT;
  return 0;

}

int init_module(void) {
  int retval;

  if((retval=check_region(io, 0x08))<0)
    return retval;
  request_region(io, 0x08, "eib");
  outb(0x00, io+UART_IER);
  retval=request_irq(irq, interrupt_handler, SA_INTERRUPT, "eib", NULL);
  if(retval) {
    release_region(io, 0x08);
    return retval;
  }
  retval=register_chrdev(major, "eib", &fops);
  if(retval<0) {
    release_region(io, 0x08);
    free_irq(irq, NULL);
    return retval;
  }
  if(major==0)
    major=retval;
  init_serial();

  memset(&eib_dev_info, 0, sizeof(struct eib_device_information));
  sprintf(eib_dev_info.string, "%s, version=%s, io=0x%x, irq=%d, major=%d", eib_name, eib_version, io, irq, major);
  printk(KERN_INFO "%s\n", eib_dev_info.string);
  eib_dev_info.rb_max_length=MAX_READ_BUFFER;
  eib_dev_info.wb_max_length=MAX_WRITE_BUFFER;
  eib_dev_info.io=io;
  eib_dev_info.irq=irq;
  eib_dev_info.major=major;
  do_gettimeofday(&last_int_tv);

  return 0;
}

void cleanup_module(void) {

#ifdef EIB_DEBUG
  printk(KERN_DEBUG "debug: cleanup_module: state=%d, warning_flag=%lx, initate_send=%d, getCTS()=%d, MCR=%x\n", state, warning_flag, initate_send, getCTS(), inb(io+UART_MCR));
#endif

  if(rb_length)
    printk(KERN_WARNING "eib: warning: read buffer was not empty.\n");

  del_timer(&initate_send_timer_list);
  del_timer(&detect_reset_timer_list);
  restore_serial();
  release_region(io, 0x08);
  free_irq(irq, NULL);
  unregister_chrdev(major, "eib");
}

void initate_send_timer(unsigned long data) {
  unsigned long flags;

  if(wb_length==0) {
    warning_flag|=0x10000; // initate send with empty write buffer
    return;
  }

  save_flags(flags); cli();
  if(state==0 && !getCTS() && time_difference(&last_int_tv, NULL)>3000) {
    initate_send=1;
    do_gettimeofday(&start_tv);
    do_gettimeofday(&last_int_tv);
    setRTS();
    add_detect_reset_timer();
    restore_flags(flags);
    return;
  }
  add_initate_send_timer();
  restore_flags(flags);
}

void detect_reset_timer(unsigned long data) {
  unsigned long flags;

  save_flags(flags); cli();
  if(state==0 && !initate_send && !getCTS()) { // idle
    restore_flags(flags);
    return;
  } else if(state==7) { // reset
    if(time_difference(&reset_tv, NULL)>MIN_RESET_WAIT_TIME) {
      do_gettimeofday(&last_int_tv);
      if(wb_length)
	add_initate_send_timer();
      rb_col=0;
      wb_col=0;
      state=0;
      restore_flags(flags);
      return;
    }
  } else { // read || write
    if(time_difference(&start_tv, NULL)>READ_WRITE_TIMEOUT)
      do_reset(0x100000); // read write timeout
    else if(time_difference(&last_int_tv, NULL)>MAX_TIME_BETWEEN_INT)
      do_reset(0x200000); // max time between int exceeded
  }
  add_detect_reset_timer();
  restore_flags(flags);
}

inline void check_interrupt_pending(void) {
  int id=inb(io+UART_IIR);

  if(~id&UART_IIR_NO_INT) {
    if((id&UART_IIR_ID)==UART_IIR_RDI)
      do_reset(0x40000); // unexpected receive data interrupt
    else if(inb(io+UART_MSR)&UART_MSR_DCTS)
      if(getCTS())
        do_reset(0x80000); // unexpected cts change to high
      // else !getCTS() -> state unchanged
    // else uninteresting modem status register change -> ignore
  }
}

void do_temt_bh(void) {
  struct timeval tv;
  unsigned long flags;
  long temp;

  if(state==7) // reset
    return;

  if(state!=3 && state!=6) {
    do_reset(0x40); // unknown state
    return;
  }

  do_gettimeofday(&tv);
  while(~inb(io+UART_LSR)&UART_LSR_TEMT && time_difference(&tv, NULL)<MAX_TEMT_WAIT)
    ;

  if(~inb(io+UART_LSR)&UART_LSR_TEMT)
    do_reset(0x1000); // exceed MAX_TEMT_WAIT

  switch(state) {

  case 3:
    eib_dev_info.stat_rx_bytes++;
    if(rb_col==(read_buffer[rb_head][0]&0x1f)+1) { // read finished
      eib_dev_info.stat_rx_data_blocks++;
      if(eib_dev_info.stat_max_rx_time_data_block<(temp=time_difference(&start_tv, NULL)))
	eib_dev_info.stat_max_rx_time_data_block=temp;
      if(read_buffer[rb_head][0]==0xa0)
	warning_flag|=0x20000; // read bcu reset
      else {
	rb_head=(rb_head+1)%MAX_READ_BUFFER;
	rb_length++;
	wake_up_interruptible(&eib_read_queue);
      }
      rb_col=0;
      save_flags(flags); cli();
      if(wb_length)
	add_initate_send_timer();
      state=0;
    } else {
      save_flags(flags); cli();
      state=1;
    }
    check_interrupt_pending();
    clrRTS();
    restore_flags(flags);
    break;

  case 6:
    do_gettimeofday(&tv);
    while(time_difference(&tv, NULL)<WRITE_TEMT_CLRRTS_DELAY)
      ;
    eib_dev_info.stat_tx_bytes++;
    if(wb_col==(write_buffer[wb_tail][0]&0x1f)+1) { // write finished
      eib_dev_info.stat_tx_data_blocks++;
      if(eib_dev_info.stat_max_tx_time_data_block<(temp=time_difference(&start_tv, NULL)))
	eib_dev_info.stat_max_tx_time_data_block=temp;
      wb_tail=(wb_tail+1)%MAX_WRITE_BUFFER;
      wb_length--;
      wake_up_interruptible(&eib_write_queue);
      wb_col=0;
      if(wb_length)
	add_initate_send_timer();
      save_flags(flags); cli();
      state=0;
      check_interrupt_pending();
      clrRTS();
    } else {
      check_interrupt_pending();
      clrRTS();
      do_gettimeofday(&tv);
      while(time_difference(&tv, NULL)<WRITE_CLRRTS_SETRTS_DELAY)
	;
      save_flags(flags); cli();
      state=4;
      check_interrupt_pending();
      setRTS();
    }
    restore_flags(flags);
    break;

  case 7: // reset
    break;

  default:
    do_reset(0x40); // unknown state
    break;
  }
  do_gettimeofday(&last_int_tv);
}

void interrupt_handler(int irq, void *dev_id, struct pt_regs *regs) {
  int tmp=0, id;

  id=inb(io+UART_IIR);
  if(~id&UART_IIR_NO_INT) {

    if(state!=0 && state!=7 && time_difference(&start_tv, NULL)>READ_WRITE_TIMEOUT)
      do_reset(0x100000); // read write timeout

    switch(state) {

    case 0: // idle
      if((id&UART_IIR_ID)==UART_IIR_MSI) {
	if(inb(io+UART_MSR)&UART_MSR_DCTS) {
	  if(getCTS()) {
	    if(initate_send) { // write
	      outb(write_buffer[wb_tail][0], io+UART_TX);
	      initate_send=0;
	      wb_col++;
	      state=5;
	      break;
	    } else { // read
	      add_detect_reset_timer();
	      if(rb_length==MAX_READ_BUFFER) {
		do_reset(0x10);  // read buffer overflow
	      } else {
		do_gettimeofday(&start_tv);
		setRTS();
		outb(0xff, io+UART_TX);
		state=2;
	      }
	    }
	  } // else !getCTS() -> state=0
	} // else uninteresting modem status register change -> ignore
      } else
	do_reset(0x2); // unexpected receive data interrupt
      break;

    case 1:
      if((id&UART_IIR_ID)==UART_IIR_MSI) {
	if(inb(io+UART_MSR)&UART_MSR_DCTS)
	  if(getCTS()) {
	    setRTS();
	    outb(0x00, io+UART_TX);
	    state=2;
	  } // else !getCTS() -> state=1
	// else uninteresting modem status register change -> ignore
      } else
	do_reset(0x200); // unexpected receive data interrupt
      break;

    case 2:
      if((id&UART_IIR_ID)==UART_IIR_RDI) {
	tmp=inb(io+UART_RX);
	read_buffer[rb_head][rb_col++]=tmp;
	if(inb(io+UART_LSR)&UART_LSR_OE) {
	  do_reset(0x80); // overrun
	  break;
	}
	if(rb_col==1 && !(((tmp&0x60)==0x20) && 
                          ((tmp&0x80)==(bitcount(tmp&0x7f)%2)*0x80))) {
	  do_reset(0x400); // parity error
	  break;
	}
	state=3;
	init_bh(IMMEDIATE_BH, do_temt_bh);
        mark_bh(IMMEDIATE_BH);
      }
      else if(inb(io+UART_MSR)&UART_MSR_DCTS)
	if(getCTS())
	  do_reset(0x20); // unexpected cts change to high
        // else !getCTS() -> ok
      // else uninteresting modem status register change
      break;

    case 3:
    case 6:
      if((id&UART_IIR_ID)==UART_IIR_RDI)
	do_reset(0x4); // unexpected receive data interrupt
      else if(inb(io+UART_MSR)&UART_MSR_DCTS)
	if(getCTS())
	  do_reset(0x8); // unexpected cts change to high
        // else !getCTS() -> state unchanged
      // else uninteresting modem status register change -> ignore
      break;

    case 4:
      if((id&UART_IIR_ID)==UART_IIR_MSI) {
	if(inb(io+UART_MSR)&UART_MSR_DCTS) 
	  if(getCTS()) {
	    outb_p(write_buffer[wb_tail][wb_col++], io+UART_TX);
	    state=5;
	  } // else !getCTS() -> state=4
	// else uninteresting modem status register change -> ignore
      } else
	do_reset(0x800); // unexpected receive data interrupt
      break;

    case 5:
      if((id&UART_IIR_ID)==UART_IIR_RDI) {
	tmp=inb(io+UART_RX);
	state=6;
	if(inb(io+UART_LSR)&UART_LSR_OE) {
	  do_reset(0x80); // overun
	  break;
	}

#ifdef SNIFF_HACK
	if(wb_col==1 && tmp!=0xff && !ignore_read_forced) {
#else
	if(wb_col==1 && tmp!=0xff) {
#endif

	  warning_flag|=0x100; // read forced
	  wb_col=0;
	  if(rb_length==MAX_READ_BUFFER) {
	    do_reset(0x4000); // read forced && read buffer full
	    break;
	  }
	  if(((tmp&0x60)!=0x20) ||
             ((tmp&0x80)!=(bitcount(tmp&0x7f)%2)*0x80)) {
	    do_reset(0x400); // parity error
	    break;
	  }
	  read_buffer[rb_head][rb_col++]=tmp;
	  state=3;
	}
	if(wb_col>1 && tmp!=0x00) {
	  do_reset(0x2000); // unexpect byte received
	  break;
	}
	init_bh(IMMEDIATE_BH, do_temt_bh);
        mark_bh(IMMEDIATE_BH);
      } else if(inb(io+UART_MSR)&UART_MSR_DCTS)
	do_reset(0x8000); // unexpected cts change
      // else uninteresting modem status register change
      break;

    case 7: // reset
      if((id&UART_IIR_ID)==UART_IIR_MSI)
	inb(io+UART_MSR);
      else
	inb(io+UART_RX);
      break;

    default:
      do_reset(0x40); // unknown state

    }
  } else
    warning_flag|=0x1; // no interrupt pending
  do_gettimeofday(&last_int_tv);
}
